package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.config.annotations.LogicComponent
import dev.fyloz.colorrecipesexplorer.dtos.MaterialDto
import dev.fyloz.colorrecipesexplorer.dtos.MaterialSaveDto
import dev.fyloz.colorrecipesexplorer.logic.files.WriteableFileLogic
import dev.fyloz.colorrecipesexplorer.model.Material
import dev.fyloz.colorrecipesexplorer.service.MaterialService

interface MaterialLogic : Logic<MaterialDto, MaterialService> {
    /** Checks if a material with the given [name] exists. */
    fun existsByName(name: String): Boolean

    /** Gets all materials that are not a mix type. */
    fun getAllNotMixType(): Collection<MaterialDto>

    /**
     * Gets all materials available for the creation of a mix for the recipe with the given [recipeId],
     * including normal materials and materials from mix types included in the said recipe.
     */
    fun getAllForMixCreation(recipeId: Long): Collection<MaterialDto>

    /**
     * Gets all materials available for updating the mix with the given [mixId],
     * including normal materials and materials from mix types included in the mix recipe
     * and excluding the material of the mix type of the said mix.
     */
    fun getAllForMixUpdate(mixId: Long): Collection<MaterialDto>

    /** Saves the given [dto]. */
    fun save(dto: MaterialSaveDto): MaterialDto

    /** Updates the given [dto]. */
    fun update(dto: MaterialSaveDto): MaterialDto

    /** Updates the quantity of the given [material] with the given [factor] and returns the updated quantity. */
    fun updateQuantity(material: MaterialDto, factor: Float): Float
}

@LogicComponent
class DefaultMaterialLogic(
    service: MaterialService,
    val recipeLogic: RecipeLogic,
    val mixLogic: MixLogic,
    val materialTypeLogic: MaterialTypeLogic,
    val fileLogic: WriteableFileLogic
) : BaseLogic<MaterialDto, MaterialService>(service, Material::class.simpleName!!), MaterialLogic {
    override fun existsByName(name: String) = service.existsByName(name, null)
    override fun getAllNotMixType() = service.getAllNotMixType()

    override fun getAllForMixCreation(recipeId: Long): Collection<MaterialDto> {
        val recipesMixTypes = recipeLogic.getById(recipeId).mixTypes

        return getAll().filter { !it.isMixType || recipesMixTypes.any { mixType -> mixType.material.id == it.id } }
    }

    override fun getAllForMixUpdate(mixId: Long): Collection<MaterialDto> {
        val mix = mixLogic.getById(mixId)
        val recipesMixTypes = mix.recipe.mixTypes

        return getAll().filter { !it.isMixType || recipesMixTypes.any { mixType -> mixType.material.id == it.id } }
            .filter { it.id != mix.mixType.material.id }
    }

    override fun save(dto: MaterialSaveDto) = save(saveDtoToDto(dto, false)).also { saveSimdutFile(dto, false) }
    override fun save(dto: MaterialDto): MaterialDto {
        throwIfNameAlreadyExists(dto.name)

        return super.save(dto)
    }

    override fun update(dto: MaterialSaveDto) = update(saveDtoToDto(dto, true)).also { saveSimdutFile(dto, true) }
    override fun update(dto: MaterialDto): MaterialDto {
        throwIfNameAlreadyExists(dto.name, dto.id)

        return super.update(dto)
    }

    override fun updateQuantity(material: MaterialDto, factor: Float): Float {
        val updatedQuantity = material.inventoryQuantity + factor
        service.updateInventoryQuantityById(material.id, updatedQuantity)

        return updatedQuantity
    }

    override fun deleteById(id: Long) {
        if (service.isUsedByMixMaterialOrMixType(id)) {
            throw cannotDeleteException("Cannot delete the material with the id '$id' because mix types and/or recipes depends on it")
        }

        val material = getById(id)
        val simdutPath = Material.getSimdutFilePath(material.name)
        if (fileLogic.exists(simdutPath)) {
            fileLogic.delete(simdutPath)
        }

        super.deleteById(id)
    }

    private fun saveDtoToDto(saveDto: MaterialSaveDto, updating: Boolean): MaterialDto {
        val isMixType = !updating || getById(saveDto.id).isMixType
        val materialType = materialTypeLogic.getById(saveDto.materialTypeId)

        return MaterialDto(saveDto.id, saveDto.name, saveDto.inventoryQuantity, isMixType, materialType, null)
    }

    private fun saveSimdutFile(dto: MaterialSaveDto, updating: Boolean) {
        val file = dto.simdutFile

        if (file != null && !file.isEmpty) {
            fileLogic.write(file, Material.getSimdutFilePath(dto.name), updating)
        }
    }

    private fun throwIfNameAlreadyExists(name: String, id: Long? = null) {
        if (service.existsByName(name, id)) {
            throw alreadyExistsException(value = name)
        }
    }
}