package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.config.annotations.LogicComponent
import dev.fyloz.colorrecipesexplorer.dtos.MixMaterialDto
import dev.fyloz.colorrecipesexplorer.exception.InvalidPositionError
import dev.fyloz.colorrecipesexplorer.exception.InvalidPositionsException
import dev.fyloz.colorrecipesexplorer.exception.RestException
import dev.fyloz.colorrecipesexplorer.model.MixMaterial
import dev.fyloz.colorrecipesexplorer.service.MixMaterialService
import dev.fyloz.colorrecipesexplorer.utils.PositionUtils
import org.springframework.http.HttpStatus

interface MixMaterialLogic : Logic<MixMaterialDto, MixMaterialService> {
    /**
     *  Validates if the given [mixMaterials]. To be valid, the position of each mix material must be greater or equals to 1 and unique in the set.
     *  There must also be no gap between the positions. Also, the quantity of the first mix material in the set must not be expressed in percentages.
     *  If any of those criteria are not met, an [InvalidGroupStepsPositionsException] will be thrown.
     */
    fun validateMixMaterials(mixMaterials: Set<MixMaterialDto>)
}

@LogicComponent
class DefaultMixMaterialLogic(service: MixMaterialService) :
    BaseLogic<MixMaterialDto, MixMaterialService>(service, MixMaterial::class.simpleName!!), MixMaterialLogic {
    override fun validateMixMaterials(mixMaterials: Set<MixMaterialDto>) {
        if (mixMaterials.isEmpty()) return

        val sortedMixMaterials = mixMaterials.sortedBy { it.position }

        try {
            PositionUtils.validate(sortedMixMaterials.map { it.position })
        } catch (ex: InvalidPositionsException) {
            throw InvalidMixMaterialsPositionsException(ex.errors)
        }

        if (sortedMixMaterials[0].material.materialType.usePercentages) {
            throw InvalidFirstMixMaterialException(sortedMixMaterials[0])
        }
    }
}

// TODO check if required
class InvalidMixMaterialsPositionsException(
    val errors: Set<InvalidPositionError>
) : RestException(
    "invalid-mixmaterial-position",
    "Invalid mix materials positions",
    HttpStatus.BAD_REQUEST,
    "The position of mix materials are invalid",
    mapOf(
        "invalidMixMaterials" to errors
    )
)

class InvalidFirstMixMaterialException(
    val mixMaterial: MixMaterialDto
) : RestException(
    "invalid-mixmaterial-first",
    "Invalid first mix material",
    HttpStatus.BAD_REQUEST,
    "The first mix material is invalid because its material must not be expressed in percents",
    mapOf(
        "mixMaterial" to mixMaterial
    )
)