package dev.fyloz.colorrecipesexplorer.rest

import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeEditUsers
import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeViewUsers
import dev.fyloz.colorrecipesexplorer.logic.users.GroupLogic
import dev.fyloz.colorrecipesexplorer.logic.users.UserLogic
import dev.fyloz.colorrecipesexplorer.model.account.*
import org.springframework.context.annotation.Profile
import org.springframework.http.MediaType
import org.springframework.security.access.prepost.PreAuthorize
import org.springframework.web.bind.annotation.*
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import javax.validation.Valid

private const val USER_CONTROLLER_PATH = "api/user"
private const val GROUP_CONTROLLER_PATH = "api/user/group"

@RestController
@RequestMapping(USER_CONTROLLER_PATH)
@Profile("!emergency")
class UserController(private val userLogic: UserLogic) {
    @GetMapping
    @PreAuthorizeViewUsers
    fun getAll() =
        ok(userLogic.getAllForOutput())

    @GetMapping("{id}")
    @PreAuthorizeViewUsers
    fun getById(@PathVariable id: Long) =
        ok(userLogic.getByIdForOutput(id))

    @PostMapping
    @PreAuthorizeEditUsers
    fun save(@Valid @RequestBody user: UserSaveDto) =
        created<UserOutputDto>(USER_CONTROLLER_PATH) {
            with(userLogic) {
                save(user).toOutput()
            }
        }

    @PutMapping
    @PreAuthorizeEditUsers
    fun update(@Valid @RequestBody user: UserUpdateDto) =
        noContent {
            userLogic.update(user)
        }

    @PutMapping("{id}/password", consumes = [MediaType.TEXT_PLAIN_VALUE])
    @PreAuthorizeEditUsers
    fun updatePassword(@PathVariable id: Long, @RequestBody password: String) =
        noContent {
            userLogic.updatePassword(id, password)
        }

    @PutMapping("{userId}/permissions/{permission}")
    @PreAuthorizeEditUsers
    fun addPermission(
        @PathVariable userId: Long,
        @PathVariable permission: Permission
    ) = noContent {
        userLogic.addPermission(userId, permission)
    }

    @DeleteMapping("{userId}/permissions/{permission}")
    @PreAuthorizeEditUsers
    fun removePermission(
        @PathVariable userId: Long,
        @PathVariable permission: Permission
    ) = noContent {
        userLogic.removePermission(userId, permission)
    }

    @DeleteMapping("{id}")
    @PreAuthorizeEditUsers
    fun deleteById(@PathVariable id: Long) =
        userLogic.deleteById(id)
}

@RestController
@RequestMapping(GROUP_CONTROLLER_PATH)
@Profile("!emergency")
class GroupsController(
    private val groupLogic: GroupLogic,
    private val userLogic: UserLogic
) {
    @GetMapping
    @PreAuthorize("hasAnyAuthority('VIEW_RECIPES', 'VIEW_USERS')")
    fun getAll() =
        ok(groupLogic.getAllForOutput())

    @GetMapping("{id}")
    @PreAuthorizeViewUsers
    fun getById(@PathVariable id: Long) =
        ok(groupLogic.getByIdForOutput(id))

    @GetMapping("{id}/users")
    @PreAuthorizeViewUsers
    fun getUsersForGroup(@PathVariable id: Long) =
        ok(with(userLogic) {
            groupLogic.getUsersForGroup(id)
                .map { it.toOutput() }
        })

    @PostMapping("default/{groupId}")
    @PreAuthorizeViewUsers
    fun setDefaultGroup(@PathVariable groupId: Long, response: HttpServletResponse) =
        noContent {
            groupLogic.setResponseDefaultGroup(groupId, response)
        }

    @GetMapping("default")
    @PreAuthorizeViewUsers
    fun getRequestDefaultGroup(request: HttpServletRequest) =
        ok(with(groupLogic) {
            getRequestDefaultGroup(request).toOutput()
        })

    @GetMapping("currentuser")
    fun getCurrentGroupUser(request: HttpServletRequest) =
        ok(with(groupLogic.getRequestDefaultGroup(request)) {
            userLogic.getDefaultGroupUser(this).toOutputDto()
        })

    @PostMapping
    @PreAuthorizeEditUsers
    fun save(@Valid @RequestBody group: GroupSaveDto) =
        created<GroupOutputDto>(GROUP_CONTROLLER_PATH) {
            with(groupLogic) {
                save(group).toOutput()
            }
        }

    @PutMapping
    @PreAuthorizeEditUsers
    fun update(@Valid @RequestBody group: GroupUpdateDto) =
        noContent {
            groupLogic.update(group)
        }

    @DeleteMapping("{id}")
    @PreAuthorizeEditUsers
    fun deleteById(@PathVariable id: Long) =
        noContent {
            groupLogic.deleteById(id)
        }
}

@RestController
@RequestMapping("api")
@Profile("!emergency")
class LogoutController(private val userLogic: UserLogic) {
    @GetMapping("logout")
    @PreAuthorize("isFullyAuthenticated()")
    fun logout(request: HttpServletRequest) =
        ok {
            userLogic.logout(request)
        }
}
