package dev.fyloz.colorrecipesexplorer.utils

import dev.fyloz.colorrecipesexplorer.model.ModelEntity

/** Returns a list containing the result of the given [transform] applied to each item of the [Iterable]. If the given [transform] throws, the [Throwable] will be passed to the given [throwableConsumer]. */
inline fun <T, R, reified E : Throwable> Iterable<T>.mapMayThrow(
    throwableConsumer: (E) -> Unit = {},
    transform: (T) -> R
): List<R> = this.mapNotNull {
    try {
        transform(it)
    } catch (th: Throwable) {
        if (th is E) {
            throwableConsumer(th)
            null
        } else {
            throw th
        }
    }
}

/** Find duplicated keys in the given [Iterable], using keys obtained from the given [keySelector]. */
inline fun <T, K> Iterable<T>.findDuplicated(keySelector: (T) -> K) =
    this.groupBy(keySelector)
        .filter { it.value.count() > 1 }
        .map { it.key }

/** Find duplicated elements in the given [Iterable]. */
fun <T> Iterable<T>.findDuplicated() =
    this.groupBy { it }
        .filter { it.value.count() > 1 }
        .map { it.key }

/** Check if the given [Iterable] has gaps between each element, using keys obtained from the given [keySelector]. */
inline fun <T> Iterable<T>.hasGaps(keySelector: (T) -> Int) =
    this.map(keySelector)
        .toIntArray()
        .sorted()
        .filterIndexed { index, it -> it != index + 1 }
        .isNotEmpty()

/** Check if the given [Int] [Iterable] has gaps between each element. */
fun Iterable<Int>.hasGaps() =
    this.sorted()
        .filterIndexed { index, it -> it != index + 1 }
        .isNotEmpty()

/** Clears and fills the given [MutableCollection] with the given [elements]. */
fun <T> MutableCollection<T>.setAll(elements: Collection<T>) {
    this.clear()
    this.addAll(elements)
}

/** Removes and returns all elements of a [MutableCollection] matching the given [predicate]. */
inline fun <T> MutableCollection<T>.excludeAll(predicate: (T) -> Boolean): Iterable<T> {
    val matching = this.filter(predicate)
    this.removeAll(matching.toSet())
    return matching
}

/** Merge to [ModelEntity] [Iterable]s and prevent id duplication. */
fun <T : ModelEntity> Iterable<T>.merge(other: Iterable<T>) =
    this
        .filter { model -> other.all { it.id != model.id } }
        .plus(other)
